"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.encodeRepo = exports.makeSchemaCodecs = exports.makeSchemaCodec = exports.clientVersionToSchemaVersion = exports.SchemaCodecVersion = void 0;
const internal_1 = require("@fluidframework/core-utils/internal");
const index_js_1 = require("../../codec/index.js");
const index_js_2 = require("../../core/index.js");
const index_js_3 = require("../../util/index.js");
const formatV1_js_1 = require("./formatV1.js");
/**
 * Versions for the codec that encodes an in-memory representation of a stored schema {@link TreeStoredSchema} into a persisted format (or decodes it in the opposite direction).
 */
var SchemaCodecVersion;
(function (SchemaCodecVersion) {
    SchemaCodecVersion[SchemaCodecVersion["v1"] = 1] = "v1";
})(SchemaCodecVersion || (exports.SchemaCodecVersion = SchemaCodecVersion = {}));
/**
 * Convert a FluidClientVersion to a SchemaCodecVersion.
 * @param clientVersion - The FluidClientVersion to convert.
 * @returns The SchemaCodecVersion that corresponds to the provided FluidClientVersion.
 */
function clientVersionToSchemaVersion(clientVersion) {
    // Only one version of the schema codec is currently supported.
    return SchemaCodecVersion.v1;
}
exports.clientVersionToSchemaVersion = clientVersionToSchemaVersion;
/**
 * Create a schema codec.
 * @param options - Specifies common codec options, including which `validator` to use.
 * @param writeVersion - The schema write version.
 * @returns The composed codec.
 *
 * @privateRemarks We should consider using the Shared Tree format version instead as it may be more valuable for application authors than the schema version.
 */
function makeSchemaCodec(options, writeVersion) {
    const family = makeSchemaCodecs(options);
    return (0, index_js_1.makeVersionDispatchingCodec)(family, { ...options, writeVersion });
}
exports.makeSchemaCodec = makeSchemaCodec;
/**
 * Create a family of schema codecs.
 * @param options - Specifies common codec options, including which `validator` to use.
 * @returns The composed codec family.
 */
function makeSchemaCodecs(options) {
    return (0, index_js_1.makeCodecFamily)([[SchemaCodecVersion.v1, makeSchemaCodecV1(options)]]);
}
exports.makeSchemaCodecs = makeSchemaCodecs;
/**
 * Encode an in-memory TreeStoredSchema into the specified format version.
 * @param repo - The in-memory schema.
 * @param version - The schema write version.
 * @returns The encoded schema.
 */
function encodeRepo(repo, version) {
    switch (version) {
        case SchemaCodecVersion.v1:
            return encodeRepoV1(repo);
        default:
            (0, internal_1.unreachableCase)(version);
    }
}
exports.encodeRepo = encodeRepo;
function encodeRepoV1(repo) {
    const nodeSchema = Object.create(null);
    const rootFieldSchema = (0, index_js_2.encodeFieldSchema)(repo.rootFieldSchema);
    for (const name of [...repo.nodeSchema.keys()].sort()) {
        const schema = repo.nodeSchema.get(name) ?? (0, internal_1.fail)(0xb28 /* missing schema */);
        Object.defineProperty(nodeSchema, name, {
            enumerable: true,
            configurable: true,
            writable: true,
            value: schema.encode(),
        });
    }
    return {
        version: index_js_2.schemaFormatV1.version,
        nodes: nodeSchema,
        root: rootFieldSchema,
    };
}
function decode(f) {
    const nodeSchema = new Map();
    for (const [key, schema] of Object.entries(f.nodes)) {
        nodeSchema.set((0, index_js_3.brand)(key), index_js_2.storedSchemaDecodeDispatcher.dispatch(schema));
    }
    return {
        rootFieldSchema: (0, index_js_2.decodeFieldSchema)(f.root),
        nodeSchema,
    };
}
/**
 * Creates a codec which performs synchronous monolithic encoding of schema content.
 * @param options - Specifies common codec options, including which `validator` to use.
 * @returns The codec.
 */
function makeSchemaCodecV1(options) {
    return (0, index_js_1.makeVersionedValidatedCodec)(options, new Set([index_js_2.schemaFormatV1.version]), formatV1_js_1.Format, {
        encode: (data) => encodeRepoV1(data),
        decode: (data) => decode(data),
    });
}
//# sourceMappingURL=codec.js.map