/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { fail, unreachableCase } from "@fluidframework/core-utils/internal";
import { makeCodecFamily, makeVersionDispatchingCodec, makeVersionedValidatedCodec, } from "../../codec/index.js";
import { decodeFieldSchema, encodeFieldSchema, schemaFormatV1, storedSchemaDecodeDispatcher, } from "../../core/index.js";
import { brand } from "../../util/index.js";
import { Format as FormatV1 } from "./formatV1.js";
/**
 * Versions for the codec that encodes an in-memory representation of a stored schema {@link TreeStoredSchema} into a persisted format (or decodes it in the opposite direction).
 */
export var SchemaCodecVersion;
(function (SchemaCodecVersion) {
    SchemaCodecVersion[SchemaCodecVersion["v1"] = 1] = "v1";
})(SchemaCodecVersion || (SchemaCodecVersion = {}));
/**
 * Convert a FluidClientVersion to a SchemaCodecVersion.
 * @param clientVersion - The FluidClientVersion to convert.
 * @returns The SchemaCodecVersion that corresponds to the provided FluidClientVersion.
 */
export function clientVersionToSchemaVersion(clientVersion) {
    // Only one version of the schema codec is currently supported.
    return SchemaCodecVersion.v1;
}
/**
 * Create a schema codec.
 * @param options - Specifies common codec options, including which `validator` to use.
 * @param writeVersion - The schema write version.
 * @returns The composed codec.
 *
 * @privateRemarks We should consider using the Shared Tree format version instead as it may be more valuable for application authors than the schema version.
 */
export function makeSchemaCodec(options, writeVersion) {
    const family = makeSchemaCodecs(options);
    return makeVersionDispatchingCodec(family, { ...options, writeVersion });
}
/**
 * Create a family of schema codecs.
 * @param options - Specifies common codec options, including which `validator` to use.
 * @returns The composed codec family.
 */
export function makeSchemaCodecs(options) {
    return makeCodecFamily([[SchemaCodecVersion.v1, makeSchemaCodecV1(options)]]);
}
/**
 * Encode an in-memory TreeStoredSchema into the specified format version.
 * @param repo - The in-memory schema.
 * @param version - The schema write version.
 * @returns The encoded schema.
 */
export function encodeRepo(repo, version) {
    switch (version) {
        case SchemaCodecVersion.v1:
            return encodeRepoV1(repo);
        default:
            unreachableCase(version);
    }
}
function encodeRepoV1(repo) {
    const nodeSchema = Object.create(null);
    const rootFieldSchema = encodeFieldSchema(repo.rootFieldSchema);
    for (const name of [...repo.nodeSchema.keys()].sort()) {
        const schema = repo.nodeSchema.get(name) ?? fail(0xb28 /* missing schema */);
        Object.defineProperty(nodeSchema, name, {
            enumerable: true,
            configurable: true,
            writable: true,
            value: schema.encode(),
        });
    }
    return {
        version: schemaFormatV1.version,
        nodes: nodeSchema,
        root: rootFieldSchema,
    };
}
function decode(f) {
    const nodeSchema = new Map();
    for (const [key, schema] of Object.entries(f.nodes)) {
        nodeSchema.set(brand(key), storedSchemaDecodeDispatcher.dispatch(schema));
    }
    return {
        rootFieldSchema: decodeFieldSchema(f.root),
        nodeSchema,
    };
}
/**
 * Creates a codec which performs synchronous monolithic encoding of schema content.
 * @param options - Specifies common codec options, including which `validator` to use.
 * @returns The codec.
 */
function makeSchemaCodecV1(options) {
    return makeVersionedValidatedCodec(options, new Set([schemaFormatV1.version]), FormatV1, {
        encode: (data) => encodeRepoV1(data),
        decode: (data) => decode(data),
    });
}
//# sourceMappingURL=codec.js.map