/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { makeSharedObjectKind, } from "@fluidframework/shared-object-base/internal";
import { SharedTreeKernel, } from "./shared-tree/index.js";
import { Breakable } from "./util/index.js";
import { UsageError } from "@fluidframework/telemetry-utils/internal";
import { SharedTreeFactoryType, SharedTreeAttributes } from "./sharedTreeAttributes.js";
/**
 * Creates a factory for shared tree kernels with the given options.
 * @remarks
 * Exposes {@link ITreePrivate} to allow access to internals in tests without a cast.
 * Code exposing this beyond this package will need to update to a more public type.
 */
function treeKernelFactory(options) {
    function treeFromKernelArgs(args) {
        if (args.idCompressor === undefined) {
            throw new UsageError("IdCompressor must be enabled to use SharedTree");
        }
        return new SharedTreeKernel(new Breakable("SharedTree"), args.sharedObject, args.serializer, args.submitLocalMessage, args.lastSequenceNumber, args.logger, args.idCompressor, options);
    }
    return {
        create: (args) => {
            const k = treeFromKernelArgs(args);
            return { kernel: k, view: k.view };
        },
        async loadCore(args, storage) {
            const k = treeFromKernelArgs(args);
            await k.loadCore(storage);
            return { kernel: k, view: k.view };
        },
    };
}
/**
 * SharedTree is a hierarchical data structure for collaboratively editing strongly typed JSON-like trees
 * of objects, arrays, and other data types.
 * @legacy
 * @alpha
 */
export const SharedTree = configuredSharedTree({});
/**
 * {@link SharedTree} but allowing a non-default configuration.
 * @remarks
 * This is useful for debugging and testing to opt into extra validation or see if opting out of some optimizations fixes an issue.
 * @example
 * ```typescript
 * import {
 * 	ForestType,
 * 	TreeCompressionStrategy,
 * 	configuredSharedTree,
 * 	typeboxValidator,
 * 	// eslint-disable-next-line import/no-internal-modules
 * } from "@fluidframework/tree/internal";
 * const SharedTree = configuredSharedTree({
 * 	forest: ForestType.Reference,
 * 	jsonValidator: typeboxValidator,
 * 	treeEncodeType: TreeCompressionStrategy.Uncompressed,
 * });
 * ```
 * @privateRemarks
 * This should be legacy, but has to be internal due to limitations of API tagging preventing it from being both alpha and alpha+legacy.
 * TODO:
 * Expose Ajv validator for better error message quality somehow.
 * Maybe as part of a test utils or dev-tool package?
 * @internal
 */
export function configuredSharedTree(options) {
    const sharedObjectOptions = {
        type: SharedTreeFactoryType,
        attributes: SharedTreeAttributes,
        telemetryContextPrefix: "fluid_sharedTree_",
        factory: treeKernelFactory(options),
    };
    return makeSharedObjectKind(sharedObjectOptions);
}
//# sourceMappingURL=treeFactory.js.map